import hashlib
import math
import functools
import contextvars

import asyncio.events as events


def bin_concat(int_1, int_2, int_1_byte=None, int_2_byte=None):  # pragma: no cover
    """Concatenate two integer.
    Steps:
        1. Zero fill integers according to the given byte.
        2. Concatenate two integers in binary.

    Parameters
    ----------
    int_1 : int
        Left partition.
    int_2 : int
        Right partition.
    int_1_byte : int, optional
        If None is given, int_1 will be zero filled to the number which can be devided exactly by 8,
        otherwise int_1 will be zero filled to int_1_byte*8, by default None.
    int_2_byte : int, optional
        If None is given, int_2 will be zero filled to the number which can be devided exactly by 8,
        otherwise int_2 will be zero filled to int_2_byte*8, by default None.

    Returns
    -------
    int
        Concatenated value.
    """

    int_1_bin = bin(int_1)[2:]
    int_2_bin = bin(int_2)[2:]

    if int_1_byte is not None:
        int_1_bin = int_1_bin.zfill(int_1_byte * 8)
    else:
        int_1_bin = int_1_bin.zfill(math.ceil(len(int_1_bin) / 8) * 8)
    if int_2_byte is not None:
        int_2_bin = int_2_bin.zfill(int_2_byte * 8)
    else:
        int_2_bin = int_2_bin.zfill(math.ceil(len(int_2_bin) / 8) * 8)

    return int(int_1_bin + int_2_bin, base=2)


def cast_bytes(v, nbytes):  # pragma: no cover
    """Cast byte(s) from v.

    Parameters
    ----------
    v : int
        Value to be casted.
    nbytes : int
        Number of bytes will be casted.

    Returns
    -------
    int
        Casted value.
    """

    assert nbytes >= 1 and nbytes <= 8, "Invalid nbytes : {}".format(nbytes)
    return v & (0x100**nbytes - 1)


def convert_hash(v):  # pragma: no cover
    """Convert md5 from string.
    Steps:
        1. Use hashlib to convert md5.
        2. Perform "And 255" then "Or 256" to ensure
           the length of hashed code is limited within 511 (0b111111111).

    Parameters
    ----------
    v : str
        Value to be converted.

    Returns
    -------
    str
        Converted hash code.
    """

    md5 = hashlib.md5(v.encode("utf8")).digest()
    code = [hex((i & 0xFF) | 0x100)[3:5] for i in md5]

    return "".join(code)


def extract_bytes(v, start, end):  # pragma: no cover
    """Extract bytes scope, from start(left) to end(right).
    Steps:
        1. Right shift `end` bytes.
        2. Perform cast_bytes to extract the scope of start-end.

    Parameters
    ----------
    v : int
        Value to be extracted.
    start : int
        Starting byte.
    end : int
        Ending byte.

    Returns
    -------
    int
        Extracted value.
    """

    assert start > end and end >= 0, "Starting byte must be greater than ending byte, \
         and ending byte must be greater than zero : \
         {}, {}".format(start, end)
    return cast_bytes(v >> end * 8, start - end)



from __future__ import annotations
import asyncio
import datetime
import json
import logging
import threading
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from random import random, choices
from typing import Callable, Optional, Union

import awscrt
import httpx
from awsiot import iotshadow, mqtt_connection_builder

from .model import JciHitachiAWSStatus, JciHitachiAWSStatusSupport

AWS_REGION = "ap-northeast-1"
AWS_COGNITO_IDP_ENDPOINT = f"cognito-idp.{AWS_REGION}.amazonaws.com"
AWS_COGNITO_ENDPOINT = f"cognito-identity.{AWS_REGION}.amazonaws.com"
AWS_COGNITO_CLIENT_ID = "7kfnjsb66ei1qt5s5gjv6j1lp6"
AWS_COGNITO_USERPOOL_ID = f"{AWS_REGION}_aTZeaievK"

# AMAZON_ROOT_CERT = os.path.join(os.path.dirname(os.path.abspath(__file__)), './cert/AmazonRootCA1.pem')
AWS_IOT_ENDPOINT = "iot-api.jci-hitachi-smarthome.com"
AWS_MQTT_ENDPOINT = f"a8kcu267h96in-ats.iot.{AWS_REGION}.amazonaws.com"
QOS = awscrt.mqtt.QoS.AT_LEAST_ONCE

_LOGGER = logging.getLogger(__name__)


@dataclass
class AWSTokens:
    access_token: str
    id_token: str
    refresh_token: str
    expiration: float


@dataclass
class AWSIdentity:
    identity_id: str
    host_identity_id: str
    user_name: str
    user_attributes: dict


@dataclass
class JciHitachiMqttEvents:
    device_status: dict[str, JciHitachiAWSStatus] = field(default_factory=dict)
    device_support: dict[str, JciHitachiAWSStatusSupport] = field(default_factory=dict)
    device_control: dict[str, dict] = field(default_factory=dict)
    device_shadow: dict[str, dict] = field(default_factory=dict)
    mqtt_error: str = field(default_factory=str)
    device_status_event: dict[str, threading.Event] = field(default_factory=dict)
    device_support_event: dict[str, threading.Event] = field(default_factory=dict)
    device_control_event: dict[str, threading.Event] = field(default_factory=dict)
    device_shadow_event: dict[str, threading.Event] = field(default_factory=dict)
    mqtt_error_event: threading.Event = field(default_factory=threading.Event)


@dataclass
class JciHitachiExecutionPools:
    status_execution_pool: list = field(default_factory=list)
    shadow_execution_pool: list = field(default_factory=list)
    support_execution_pool: list = field(default_factory=list)
    control_execution_pool: list = field(default_factory=list)


class JciHitachiAWSHttpConnection(ABC):
    """Abstract class for AWS http connections."""

    @abstractmethod
    def __init__(self, print_response: bool):
        self._print_response = print_response

    @abstractmethod
    def _generate_headers(self): ...

    @abstractmethod
    def _handle_response(self, response: httpx.Response): ...

    @abstractmethod
    def _send(self): ...

    def get_data(self):
        raise NotImplementedError

    def maybe_print_http_response(self, response: httpx.Response) -> None:
        if not self._print_response:
            return

        try:
            text = json.dumps(response.json(), indent=True)
        except json.JSONDecodeError:
            text = response.text

        print("===================================================")
        print(self.__class__.__name__, "Response:")
        print("headers:", response.headers)
        print("status_code:", response.status_code)
        print("text:", text)
        print("===================================================")


class JciHitachiAWSCognitoConnection(JciHitachiAWSHttpConnection):
    """Connecting to Jci-Hitachi AWS Cognito API.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    aws_tokens : AWSTokens, optional
        If aws_tokens is given, it is used by request;
        otherwise, a login procedure is performed to obtain new aws_tokens,
        by default None.
    proxy : str, optional
        Proxy setting. Format:"schema://IP:port", e.g., http://127.0.0.1:8080, by default None.
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """

    def __init__(
        self,
        email: str,
        password: str,
        aws_tokens: Optional[AWSTokens] = None,
        proxy: Optional[str] = None,
        print_response: bool = False,
    ):
        super().__init__(print_response)
        self._login_response = None
        self._email = email
        self._password = password
        self._proxy = proxy

        if aws_tokens:
            self._aws_tokens = aws_tokens
        else:
            conn_status, self._aws_tokens = self.login()
            if conn_status != "OK":
                raise RuntimeError(
                    f"An error occurred when signing into AWS Cognito Service: {conn_status}"
                )

    def _generate_headers(self, target: str) -> dict[str, str]:
        headers = {
            "X-Amz-Target": target,
            "User-Agent": "Dalvik/2.1.0",
            "content-type": "application/x-amz-json-1.1",
            "Accept": "application/json",
        }
        return headers

    def _handle_response(self, response: httpx.Response) -> tuple(str, dict):
        response_json = response.json()

        if response.status_code == httpx.codes.ok:
            return "OK", response_json
        else:
            return (
                f"{response_json['__type']} {response_json['message']}",
                response_json,
            )

    def _send(
        self, target: str, json_data: Optional[dict] = None, endpoint: str = None
    ) -> tuple[str, dict]:
        headers = self._generate_headers(target)

        if endpoint is None:
            endpoint = (
                AWS_COGNITO_ENDPOINT
                if self.__class__.__name__ == "GetCredentials"
                else AWS_COGNITO_IDP_ENDPOINT
            )

        req = httpx.post(
            f"https://{endpoint}/",
            json=json_data,
            headers=headers,
            proxy=self._proxy,
            verify=True if self._proxy is None else False,
        )

        self.maybe_print_http_response(req)

        return self._handle_response(req)

    @property
    def aws_tokens(self) -> AWSTokens:
        return self._aws_tokens

    def login(self, use_refresh_token: bool = False) -> tuple(str, AWSTokens):
        """Login API.

        Parameters
        ----------
        use_refresh_token : bool, optional
            Whether or not to use AWSTokens.refresh_token to login.
            If AWSTokens is not provided, fallback to email and password, by default False

        Returns
        -------
        (str, AWSTokens)
            (status, aws tokens).
        """

        # https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_InitiateAuth.html
        if use_refresh_token and self._aws_tokens is not None:
            login_json_data = {
                "AuthFlow": "REFRESH_TOKEN_AUTH",
                "AuthParameters": {
                    "REFRESH_TOKEN": self._aws_tokens.refresh_token,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }
        else:
            use_refresh_token = False
            login_json_data = {
                "AuthFlow": "USER_PASSWORD_AUTH",
                "AuthParameters": {
                    "USERNAME": self._email,
                    "PASSWORD": self._password,
                },
                "ClientId": AWS_COGNITO_CLIENT_ID,
            }

        status, response = self._send(
            "AWSCognitoIdentityProviderService.InitiateAuth",
            login_json_data,
            AWS_COGNITO_IDP_ENDPOINT,
        )

        aws_tokens = None
        if status == "OK":
            auth_result = response["AuthenticationResult"]
            aws_tokens = AWSTokens(
                access_token=auth_result["AccessToken"],
                id_token=auth_result["IdToken"],
                refresh_token=self._aws_tokens.refresh_token
                if use_refresh_token
                else auth_result["RefreshToken"],
                expiration=time.time() + auth_result["ExpiresIn"],
            )
        return status, aws_tokens


class ChangePassword(JciHitachiAWSCognitoConnection):
    """API internal endpoint.
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_ChangePassword.html

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, new_password):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
            "PreviousPassword": self._password,
            "ProposedPassword": new_password,
        }

        status, response = self._send(
            "AWSCognitoIdentityProviderService.ChangePassword", json_data
        )

        return status, None


class GetUser(JciHitachiAWSCognitoConnection):
    """API internal endpoint.
    https://docs.aws.amazon.com/cognito-user-identity-pools/latest/APIReference/API_GetUser.html

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self):
        json_data = {
            "AccessToken": self._aws_tokens.access_token,
        }

        status, response = self._send(
            "AWSCognitoIdentityProviderService.GetUser", json_data
        )

        aws_identity = None
        if status == "OK":
            user_attributes = {
                attr["Name"]: attr["Value"] for attr in response["UserAttributes"]
            }
            aws_identity = AWSIdentity(
                identity_id=user_attributes["custom:cognito_identity_id"],
                host_identity_id=user_attributes["custom:host_identity_id"],
                user_name=response["Username"],
                user_attributes=user_attributes,
            )
        return status, aws_identity


class GetCredentials(JciHitachiAWSCognitoConnection):
    """API internal endpoint.

    Parameters
    ----------
    email : str
        User email.
    password : str
        User password.
    """

    def __init__(self, email, password, **kwargs):
        super().__init__(email, password, **kwargs)

    def get_data(self, aws_identity):
        json_data = {
            "IdentityId": aws_identity.identity_id,
            "Logins": {
                f"{AWS_COGNITO_IDP_ENDPOINT}/{AWS_COGNITO_USERPOOL_ID}": self._aws_tokens.id_token,
            },
        }

        status, response = self._send(
            "AWSCognitoIdentityService.GetCredentialsForIdentity", json_data
        )

        aws_credentials = None
        if status == "OK":
            aws_credentials = awscrt.auth.AwsCredentials(
                access_key_id=response["Credentials"]["AccessKeyId"],
                secret_access_key=response["Credentials"]["SecretKey"],
                session_token=response["Credentials"]["SessionToken"],
                expiration=datetime.datetime.fromtimestamp(
                    response["Credentials"]["Expiration"]
                ),
            )
        return status, aws_credentials


class JciHitachiAWSIoTConnection(JciHitachiAWSHttpConnection):
    """Connecting to Jci-Hitachi AWS IoT API.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    proxy : str, optional
        Proxy setting. Format:"schema://IP:port", e.g., http://127.0.0.1:8080, by default None.
    print_response : bool, optional
        If set, all responses of httpx will be printed, by default False.
    """

    def __init__(
        self,
        aws_tokens: AWSTokens,
        proxy: Optional[str] = None,
        print_response: bool = False,
    ):
        super().__init__(print_response)
        self._aws_tokens = aws_tokens
        self._proxy = proxy

    def _generate_headers(self, need_access_token: bool) -> dict[str, str]:
        headers = {
            "authorization": f"Bearer {self._aws_tokens.id_token}",
            "User-Agent": "Dalvik/2.1.0",
            "content-type": "application/json",
            "Accept": "application/json",
        }
        if need_access_token:
            headers["accesstoken"] = f"Bearer {self._aws_tokens.access_token}"
        return headers

    def _handle_response(self, response: httpx.Response) -> tuple[int, str, dict]:
        response_json = response.json()
        if response.status_code == httpx.codes.ok:
            code = response_json["status"]["code"]
            if code == 0:
                return code, "OK", response_json
            elif code == 6:
                return code, "Invalid email or password", response_json
            elif code == 12:
                return code, "Invalid session token", response_json
            else:
                return code, "Unknown error", response_json
        else:
            return (
                response.status_code,
                f"HTTP exception {response.status_code}",
                response_json,
            )

    def _send(
        self, api_name: str, json: Optional[dict] = None, need_access_token: bool = True
    ) -> tuple[str, dict]:
        req = httpx.post(
            f"https://{AWS_IOT_ENDPOINT}{api_name}",
            headers=self._generate_headers(need_access_token),
            json=json,
            proxy=self._proxy,
            verify=True if self._proxy is None else False,
        )

        self.maybe_print_http_response(req)

        code, message, response_json = self._handle_response(req)

        return message, response_json


class GetAllDevice(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllDevice", need_access_token=False)


class GetAllGroup(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllGroup")


class GetAllRegion(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/GetAllRegion", need_access_token=False)


class GetAvailableAggregationMonthlyData(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(
        self, thing_name: str, time_start: int, time_end: int
    ) -> tuple[str, dict]:
        json_data = {
            "ThingName": thing_name,
            "TimeStart": time_start,
            "TimeEnd": time_end,
        }
        return self._send("/GetAvailableAggregationMonthlyData", json_data)


class GetHistoryEventByUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self, time_start: int, time_end: int) -> tuple[str, dict]:
        json_data = {
            "TimeStart": time_start,
            "TimeEnd": time_end,
        }
        return self._send("/GetHistoryEventByUser", json_data)


class ListSubUser(JciHitachiAWSIoTConnection):
    """API internal endpoint.

    Parameters
    ----------
    aws_tokens : AWSTokens
        AWS tokens.
    """

    def __init__(self, aws_tokens: AWSTokens, **kwargs):
        super().__init__(aws_tokens, **kwargs)

    def get_data(self) -> tuple[str, dict]:
        return self._send("/ListSubUser")


class JciHitachiAWSMqttConnection:
    """Connecting to Jci-Hitachi AWS MQTT to get latest events.

    Parameters
    ----------
    get_credentials_callable : Callable
        Callable which takes no arguments and returns AwsCredentials.
    print_response : bool, optional
        If set, all responses of MQTT will be printed, by default False.
    """

    def __init__(
        self, get_credentials_callable: Callable, print_response: bool = False
    ):
        self._get_credentials_callable: Callable = get_credentials_callable
        self._print_response: bool = print_response

        self._mqttc: Optional[awscrt.mqtt.Connection] = None
        self._shadow_mqttc: Optional[iotshadow.IotShadowClient] = None
        self._client_tokens: dict[str, str] = {}
        self._mqtt_events: JciHitachiMqttEvents = JciHitachiMqttEvents()
        self._execution_lock: threading.Lock = threading.Lock()
        self._execution_pools: JciHitachiExecutionPools = JciHitachiExecutionPools()

    def __del__(self):
        self.disconnect()

    @property
    def mqtt_events(self) -> JciHitachiMqttEvents:
        """MQTT events.

        Returns
        -------
        JciHitachiMqttEvents
            See JciHitachiMqttEvents.
        """

        return self._mqtt_events

    def _on_publish(self, topic: str, payload: bytes, dup, qos, retain, **kwargs):
        try:
            payload = json.loads(payload.decode())
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error(
                f"Mqtt topic {topic} published with payload {payload} cannot be decoded: {e}"
            )
            return

        if self._print_response:
            print(f"Mqtt topic {topic} published with payload \n {payload}")

        split_topic = topic.split("/")

        if len(split_topic) >= 4 and split_topic[3] != "shadow":
            thing_name = split_topic[1]
            if split_topic[2] == "status" and split_topic[3] == "response":
                self._mqtt_events.device_status[thing_name] = JciHitachiAWSStatus(
                    payload
                )
                self._mqtt_events.device_status_event[thing_name].set()
            elif split_topic[2] == "registration" and split_topic[3] == "response":
                self._mqtt_events.device_support[thing_name] = (
                    JciHitachiAWSStatusSupport(payload)
                )
                self._mqtt_events.device_support_event[thing_name].set()
            elif split_topic[2] == "control" and split_topic[3] == "response":
                self._mqtt_events.device_control[thing_name] = payload
                self._mqtt_events.device_control_event[thing_name].set()

    def _on_update_named_shadow_accepted(self, response):
        try:
            thing_name = self._client_tokens.pop(response.client_token)
        except:
            _LOGGER.error(
                f"An unknown shadow response is received. Client token: {response.client_token}"
            )
            return

        if self._print_response:
            print(f"An `update` shadow response is received: {response.state.reported}")

        if response.state:
            if response.state.reported:
                self._mqtt_events.device_control[thing_name] = response.state.reported
                self._mqtt_events.device_control_event[thing_name].set()

    def _on_update_named_shadow_rejected(self, error):
        _LOGGER.error(
            f"A shadow request {error.client_token} was rejected by the API: {error.code} {error.message}"
        )

    def _on_get_named_shadow_accepted(self, response):
        try:
            thing_name = self._client_tokens.pop(response.client_token)
        except:
            _LOGGER.error(
                f"An unknown shadow response is received. Client token: {response.client_token}"
            )
            return

        if self._print_response:
            print(f"A `get` shadow response is received: {response.state.reported}")

        if response.state:
            if response.state.reported:
                self._mqtt_events.device_shadow[thing_name] = response.state.reported
                self._mqtt_events.device_shadow_event[thing_name].set()

    def _on_get_named_shadow_rejected(self, error):
        _LOGGER.error(
            f"A shadow request {error.client_token} was rejected by the API: {error.code} {error.message}"
        )

    def _on_message(self, topic, payload, dup, qos, retain, **kwargs):
        return

    def _on_connection_interrupted(self, connection, error, **kwargs):
        _LOGGER.error(f"MQTT connection was interrupted with exception {error}")
        self._mqtt_events.mqtt_error = error.__class__.__name__
        self._mqtt_events.mqtt_error_event.set()

    def _on_connection_resumed(
        self, connection, return_code, session_present, **kwargs
    ):
        if session_present:
            _LOGGER.info("MQTT connection was resumed.")
        else:
            _LOGGER.info(
                "MQTT connection was resumed, but the previous session was lost. Resubscribing..."
            )
            resubscribe_future, packet_id = connection.resubscribe_existing_topics()

            def on_resubscribe_complete(resubscribe_future):
                try:
                    resubscribe_results = resubscribe_future.result()
                    assert resubscribe_results["packet_id"] == packet_id
                    for topic, qos in resubscribe_results["topics"]:
                        assert qos is not None
                except Exception as e:
                    _LOGGER.error("Resubscribe failure:", e)

            resubscribe_future.add_done_callback(on_resubscribe_complete)
            _LOGGER.info("Resubscribed successfully.")
        return

    async def _wrap_async(self, identifier: str, fn: Callable) -> str:
        await asyncio.sleep(
            random() / 2
        )  # randomly wait 0~0.5 seconds to prevent messages flooding to the broker.
        await asyncio.to_thread(fn)
        return identifier

    def disconnect(self) -> None:
        """Disconnect from the MQTT broker."""

        if self._mqttc is not None:
            self._mqttc.disconnect()

    def configure(self, identity_id) -> None:
        """Configure MQTT.

        Parameters
        ----------
        identity_id : str
            Identity ID.
        """

        cred_provider = awscrt.auth.AwsCredentialsProvider.new_delegate(
            self._get_credentials_callable
        )
        event_loop_group = awscrt.io.EventLoopGroup(1)
        host_resolver = awscrt.io.DefaultHostResolver(event_loop_group)
        client_bootstrap = awscrt.io.ClientBootstrap(event_loop_group, host_resolver)
        self._mqttc = mqtt_connection_builder.websockets_with_default_aws_signing(
            AWS_REGION,
            cred_provider,
            client_bootstrap=client_bootstrap,
            endpoint=AWS_MQTT_ENDPOINT,
            client_id=f"{identity_id}_{''.join(choices('abcdef0123456789', k=16))}",  # {identityid}_{64bit_hex}
            on_connection_interrupted=self._on_connection_interrupted,
            on_connection_resumed=self._on_connection_resumed,
        )
        self._mqttc.on_message(self._on_message)
        self._shadow_mqttc = iotshadow.IotShadowClient(self._mqttc)

    def connect(
        self,
        host_identity_id: str,
        shadow_names: Optional[Union[str, list[str]]] = None,
        thing_names: Optional[Union[str, list[str]]] = None,
    ) -> bool:
        """Connect to the MQTT broker and start loop.

        Parameters
        ----------
        host_identity_id : str
            Host identity ID.
        shadow_names : str or list of str, optional
            Names to be subscribed in Shadow, by default None.
        thing_names : str or list of str, optional
            Things to be subscribed in Shadow, by default None.

        Returns
        -------
        bool
            A bool indicating whether the mqtt is successfully connected and subscribed.
        """

        try:
            connect_future = self._mqttc.connect()
            connect_future.result()
            _LOGGER.info("MQTT Connected.")
        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            _LOGGER.error("MQTT connection failed with exception {}".format(e))
            return False

        try:
            subscribe_future, _ = self._mqttc.subscribe(
                f"{host_identity_id}/+/+/response", QOS, callback=self._on_publish
            )
            subscribe_future.result()

            if thing_names is not None and shadow_names is not None:
                shadow_names = (
                    [shadow_names] if isinstance(shadow_names, str) else shadow_names
                )
                thing_names = (
                    [thing_names] if isinstance(thing_names, str) else thing_names
                )

                for shadow_name in shadow_names:
                    for thing_name in thing_names:
                        (
                            update_accepted_subscribed_future,
                            _,
                        ) = self._shadow_mqttc.subscribe_to_update_named_shadow_accepted(
                            request=iotshadow.UpdateNamedShadowSubscriptionRequest(
                                shadow_name=shadow_name, thing_name=thing_name
                            ),
                            qos=QOS,
                            callback=self._on_update_named_shadow_accepted,
                        )

                        (
                            update_rejected_subscribed_future,
                            _,
                        ) = self._shadow_mqttc.subscribe_to_update_named_shadow_rejected(
                            request=iotshadow.UpdateNamedShadowSubscriptionRequest(
                                shadow_name=shadow_name, thing_name=thing_name
                            ),
                            qos=QOS,
                            callback=self._on_update_named_shadow_rejected,
                        )

                        # Wait for subscriptions to succeed
                        update_accepted_subscribed_future.result()
                        update_rejected_subscribed_future.result()

                        (
                            get_accepted_subscribed_future,
                            _,
                        ) = self._shadow_mqttc.subscribe_to_get_named_shadow_accepted(
                            request=iotshadow.GetNamedShadowSubscriptionRequest(
                                shadow_name=shadow_name, thing_name=thing_name
                            ),
                            qos=QOS,
                            callback=self._on_get_named_shadow_accepted,
                        )

                        (
                            get_rejected_subscribed_future,
                            _,
                        ) = self._shadow_mqttc.subscribe_to_get_named_shadow_rejected(
                            request=iotshadow.GetNamedShadowSubscriptionRequest(
                                shadow_name=shadow_name, thing_name=thing_name
                            ),
                            qos=QOS,
                            callback=self._on_get_named_shadow_rejected,
                        )

                        # Wait for subscriptions to succeed
                        get_accepted_subscribed_future.result()
                        get_rejected_subscribed_future.result()

        except Exception as e:
            self._mqtt_events.mqtt_error = e.__class__.__name__
            self._mqtt_events.mqtt_error_event.set()
            self.disconnect()
            _LOGGER.error("MQTT subscription failed with exception {}".format(e))
            return False
        return True

    def publish(
        self,
        host_identity_id: str,
        thing_name: str,
        publish_type: str,
        timeout: float = 10.0,
        payload: Optional[dict] = None,
    ) -> None:
        """Put messages to be published in the execution pool. execute() should be called to start async publish.

        Parameters
        ----------
        host_identity_id : str
            Host identity id.
        thing_name : str
            Thing name.
        publish_type: str
            Publish type. There are three types available: `support`, `status`, and `control`.
        timeout: float, optional
            Timeout for messages published, by default 10.0.
        payload : dict, optional
            Payload to publish, by default None.
        """

        default_payload = {"Timestamp": time.time()}

        if publish_type == "support":
            support_topic = f"{host_identity_id}/{thing_name}/registration/request"
            if thing_name in self._mqtt_events.device_support_event:
                self._mqtt_events.device_support_event[thing_name].clear()
            else:
                self._mqtt_events.device_support_event[thing_name] = threading.Event()

            def fn():
                publish_future, _ = self._mqttc.publish(
                    support_topic, json.dumps(default_payload), QOS
                )
                publish_future.result(timeout)
                self._mqtt_events.device_support_event[thing_name].wait(timeout)

            self._execution_pools.support_execution_pool.append(
                self._wrap_async(thing_name, fn)
            )
        elif publish_type == "status":
            status_topic = f"{host_identity_id}/{thing_name}/status/request"
            if thing_name in self._mqtt_events.device_status_event:
                self._mqtt_events.device_status_event[thing_name].clear()
            else:
                self._mqtt_events.device_status_event[thing_name] = threading.Event()

            def fn():
                publish_future, _ = self._mqttc.publish(
                    status_topic, json.dumps(default_payload), QOS
                )
                publish_future.result(timeout)
                self._mqtt_events.device_status_event[thing_name].wait(timeout)

            self._execution_pools.status_execution_pool.append(
                self._wrap_async(thing_name, fn)
            )
        elif publish_type == "control":
            control_topic = f"{host_identity_id}/{thing_name}/control/request"
            if thing_name in self._mqtt_events.device_control_event:
                self._mqtt_events.device_control_event[thing_name].clear()
            else:
                self._mqtt_events.device_control_event[thing_name] = threading.Event()

            def fn():
                publish_future, _ = self._mqttc.publish(
                    control_topic, json.dumps(payload), QOS
                )
                publish_future.result(timeout)
                self._mqtt_events.device_control_event[thing_name].wait(timeout)

            self._execution_pools.control_execution_pool.append(
                self._wrap_async(thing_name, fn)
            )

        else:
            raise ValueError(f"Invalid publish_type: {publish_type}")

    def publish_shadow(
        self,
        thing_name: str,
        command_name: str,
        payload: dict = {},
        shadow_name: Optional[str] = None,
        timeout: float = 10.0,
    ) -> None:
        """Publish message to IoT Shadow Service.

        Parameters
        ----------
        thing_name : str
            Thing name.
        command_name : str
            Command name, which can be `get` or `update`.
        payload : dict, optional
            Payload to publish, by default {}.
        shadow_name : str, optional
            Shadow name, by default None.
        timeout: float, optional
            Timeout for messages published, by default 10.0.
        """

        if command_name not in ["get", "update"]:  # we don't subscribe delete
            raise ValueError("command_name must be one of `get` or `update`.")

        # The length of client token can't exceed 64 bytes, so we only use gateway mac address as the token.
        client_token = thing_name.split("_")[1]
        self._client_tokens.update({client_token: thing_name})
        if thing_name in self._mqtt_events.device_shadow_event:
            self._mqtt_events.device_shadow_event[thing_name].clear()
        else:
            self._mqtt_events.device_shadow_event[thing_name] = threading.Event()

        def fn():
            if shadow_name is None:
                if command_name == "get":
                    publish_future = self._shadow_mqttc.publish_get_shadow(
                        iotshadow.GetShadowRequest(
                            client_token=client_token, thing_name=thing_name
                        ),
                        qos=QOS,
                    )
                elif command_name == "update":
                    publish_future = self._shadow_mqttc.publish_update_shadow(
                        iotshadow.UpdateShadowRequest(
                            client_token=client_token,
                            state=iotshadow.ShadowState(reported=payload),
                            thing_name=thing_name,
                        ),
                        qos=QOS,
                    )
                elif command_name == "delete":
                    publish_future = self._shadow_mqttc.publish_delete_shadow(
                        iotshadow.DeleteShadowRequest(
                            client_token=client_token, thing_name=thing_name
                        ),
                        qos=QOS,
                    )

            else:
                if command_name == "get":
                    publish_future = self._shadow_mqttc.publish_get_named_shadow(
                        iotshadow.GetNamedShadowRequest(
                            client_token=client_token,
                            shadow_name=shadow_name,
                            thing_name=thing_name,
                        ),
                        qos=QOS,
                    )
                elif command_name == "update":
                    publish_future = self._shadow_mqttc.publish_update_named_shadow(
                        iotshadow.UpdateNamedShadowRequest(
                            client_token=client_token,
                            shadow_name=shadow_name,
                            state=iotshadow.ShadowState(reported=payload),
                            thing_name=thing_name,
                        ),
                        qos=QOS,
                    )
                elif command_name == "delete":
                    publish_future = self._shadow_mqttc.publish_delete_named_shadow(
                        iotshadow.DeleteNamedShadowRequest(
                            client_token=client_token,
                            shadow_name=shadow_name,
                            thing_name=thing_name,
                        ),
                        qos=QOS,
                    )
            publish_future.result(timeout)
            self._mqtt_events.device_shadow_event[thing_name].wait(timeout)

        self._execution_pools.shadow_execution_pool.append(
            self._wrap_async(thing_name, fn)
        )

    def execute(
        self, control: bool = False
    ) -> list[
        list[Union[str, BaseException]],
        list[Union[str, BaseException]],
        list[Union[str, BaseException]],
        list[Union[str, BaseException]],
    ]:
        """Execute publish commands in the execution pools.

        Parameters
        ----------
        control : bool
            If True, commands in the `control_execution_pool` will be executed; otherwise, commands in other execution pools will be executed.

        Returns
        -------
        list
            Execution results of support, shadow, status, control, respectively.
            Each result is a list containing thing names if the execution was successful or BaseException(s) if an error occurred during execution.
        """

        async def runner():
            a, b, c, d = None, None, None, None
            if control and len(self._execution_pools.control_execution_pool) != 0:
                d = await asyncio.gather(
                    *self._execution_pools.control_execution_pool,
                    return_exceptions=True,
                )
                self._execution_pools.control_execution_pool.clear()
            else:
                if len(self._execution_pools.support_execution_pool) != 0:
                    a = await asyncio.gather(
                        *self._execution_pools.support_execution_pool,
                        return_exceptions=True,
                    )
                    self._execution_pools.support_execution_pool.clear()
                if len(self._execution_pools.shadow_execution_pool) != 0:
                    b = await asyncio.gather(
                        *self._execution_pools.shadow_execution_pool,
                        return_exceptions=True,
                    )
                    self._execution_pools.shadow_execution_pool.clear()
                if len(self._execution_pools.status_execution_pool) != 0:
                    c = await asyncio.gather(
                        *self._execution_pools.status_execution_pool,
                        return_exceptions=True,
                    )
                    self._execution_pools.status_execution_pool.clear()

            return a, b, c, d

        locked = self._execution_lock.locked()
        if locked:
            _LOGGER.debug("Other execution in progress, waiting for a lock.")

        with self._execution_lock:
            if locked:
                _LOGGER.debug("Lock acquired.")
            results = asyncio.run(runner())

        return results



import time
from unittest.mock import MagicMock, patch

import pytest

from JciHitachi.api import AWSThing, JciHitachiAWSAPI
from JciHitachi.aws_connection import AWSTokens, AWSIdentity
from JciHitachi.model import JciHitachiAWSStatus, JciHitachiAWSStatusSupport

from . import MOCK_GATEWAY_MAC, MOCK_DEVICE_AC, MOCK_DEVICE_DH, MOCK_DEVICE_HE


MOCK_THINGS_JSON = {
    "results": {
        "Things": [
            {
                "DeviceType": "1",
                "ThingName": f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}",
                "CustomDeviceName": MOCK_DEVICE_AC,
            },
            {
                "DeviceType": "2",
                "ThingName": f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}",
                "CustomDeviceName": MOCK_DEVICE_DH,
            },
        ]
    }
}


@pytest.fixture()
def fixture_aws_mock_ac_thing():
    thing = AWSThing(
        {
            "DeviceType": "1",
            "ThingName": f"ap-northeast-1:8916b515-8394-4ccd-95b8-4f553c13dafa_{MOCK_GATEWAY_MAC}",
            "CustomDeviceName": MOCK_DEVICE_AC,
        }
    )
    thing.status_code = JciHitachiAWSStatus(
        {
            "DeviceType": 1,
            "FanSpeed": 4,  # high
            "TemperatureSetting": 26,
        }
    )
    thing.support_code = JciHitachiAWSStatusSupport(
        {
            "DeviceType": 1,
            "Model": "RAD-90NF",
            "FirmwareVersion": "6.0.035",
            "FanSpeed": 31,  # 0b11111
            "TemperatureSetting": 4128,  # 32 16
        }
    )
    return thing


@pytest.fixture()
def fixture_aws_mock_dh_thing():
    thing = AWSThing(
        {
            "DeviceType": "2",
            "ThingName": f"ap-northeast-1:9c8c1d20-b0d1-11ec-9f5a-644bf019ccc9_{MOCK_GATEWAY_MAC}",
            "CustomDeviceName": MOCK_DEVICE_DH,
        }
    )
    thing.status_code = JciHitachiAWSStatus(
        {
            "DeviceType": 2,
            "Mode": 4,  # air_purify
        }
    )
    thing.support_code = JciHitachiAWSStatusSupport(
        {
            "DeviceType": 2,
            "Mode": 31,  # 0b11111
            "FirmwareVersion": "6.0.035",
            "Model": "RD-360HH",
        }
    )
    return thing


@pytest.fixture()
def fixture_aws_mock_he_thing():
    thing = AWSThing(
        {
            "DeviceType": "3",
            "ThingName": f"ap-northeast-1:667bf9f1-9671-469b-adf3-935e4c982dc6_{MOCK_GATEWAY_MAC}",
            "CustomDeviceName": MOCK_DEVICE_HE,
        }
    )
    thing.status_code = JciHitachiAWSStatus(
        {
            "DeviceType": 3,
            "BreathMode": 2,  # normal
        }
    )
    thing.support_code = JciHitachiAWSStatusSupport(
        {
            "DeviceType": 3,
            "BreathMode": 7,  # 0b111
            "FirmwareVersion": "6.0.036",
            "Model": "KPI-H",
        }
    )
    return thing


@pytest.fixture()
def fixture_aws_mock_api(
    fixture_aws_mock_ac_thing, fixture_aws_mock_dh_thing, fixture_aws_mock_he_thing
):
    api = JciHitachiAWSAPI("", "")
    api._aws_tokens = AWSTokens("", "", "", time.time() + 3600)
    api._things = {
        MOCK_DEVICE_AC: fixture_aws_mock_ac_thing,
        MOCK_DEVICE_DH: fixture_aws_mock_dh_thing,
        MOCK_DEVICE_HE: fixture_aws_mock_he_thing,
    }
    return api


@pytest.fixture()
def fixture_aws_identity():
    # dummy identity
    return AWSIdentity("id", "host_id", "username", {"attr", "attr"})


class TestAWSAPI:
    def test_login(self, fixture_aws_mock_api, fixture_aws_identity):
        api = fixture_aws_mock_api
        with patch(
            "JciHitachi.aws_connection.GetUser.get_data"
        ) as mock_get_data_get_user, patch(
            "JciHitachi.aws_connection.GetUser.login"
        ) as mock_login_get_user, patch(
            "JciHitachi.aws_connection.ListSubUser.get_data"
        ) as mock_get_data_list_subuser, patch(
            "JciHitachi.aws_connection.GetAllDevice.get_data"
        ) as mock_get_data_get_all_device, patch(
            "JciHitachi.aws_connection.JciHitachiAWSMqttConnection.connect"
        ) as mock_connect_mqtt:
            # initialization
            aws_tokens = api._aws_tokens
            api._aws_tokens = None
            mock_get_data_get_user.return_value = (
                "OK",
                fixture_aws_identity,
            )
            mock_login_get_user.return_value = ("OK", aws_tokens)

            # mock_get_data_list_subuser currently unused.
            mock_get_data_list_subuser.return_value = (
                "OK",
                {
                    "results": {
                        "FamilyMemberList": [
                            {"userId": "uid1", "isHost": False},
                            {"userId": "uid2", "isHost": True},
                        ]
                    }
                },
            )
            mock_get_data_get_all_device.return_value = ("OK", MOCK_THINGS_JSON)
            mock_connect_mqtt.return_value = True
            api.refresh_status = MagicMock()
            assert api._aws_tokens is None
            assert api._aws_identity is None
            assert len(api._things) == 3
            assert api.device_names is None

            # normal case
            api.login()
            assert api._aws_tokens == aws_tokens
            assert api._aws_identity is not None
            assert api._aws_identity.identity_id == "id"
            assert api._aws_identity.host_identity_id == "host_id"
            assert len(api._things) == 2
            assert len(api.device_names) == 2

    def test_change_password(self, fixture_aws_mock_api):
        api = fixture_aws_mock_api
        with patch(
            "JciHitachi.aws_connection.ChangePassword.get_data"
        ) as mock_get_data_1, patch(
            "JciHitachi.connection.UpdateUserCredential.get_data"
        ) as mock_get_data_2:
            mock_get_data_1.return_value = ("OK", "")
            mock_get_data_2.return_value = ("OK", "")
            api.change_password("new_password")

            mock_get_data_1.return_value = ("Not OK", "")
            with pytest.raises(
                RuntimeError,
                match="An error occurred when changing AWS Cognito password: Not OK",
            ):
                api.change_password("new_password")

            mock_get_data_1.return_value = ("OK", "")
            mock_get_data_2.return_value = ("Not OK", "")
            with pytest.raises(
                RuntimeError,
                match="An error occurred when changing Hitachi password: Not OK",
            ):
                api.change_password("new_password")

    def test_get_status(self, fixture_aws_mock_api):
        # Test status
        api = fixture_aws_mock_api
        statuses = api.get_status()
        assert isinstance(statuses[MOCK_DEVICE_AC], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_AC].FanSpeed == "high"
        assert statuses[MOCK_DEVICE_AC].status == {
            "DeviceType": "AC",
            "FanSpeed": "high",
            "TemperatureSetting": 26,
            "max_temp": 32,
            "min_temp": 16,
        }
        assert isinstance(statuses[MOCK_DEVICE_DH], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_DH].Mode == "air_purify"
        assert statuses[MOCK_DEVICE_DH].status == {
            "DeviceType": "DH",
            "Mode": "air_purify",
            "max_humidity": 70,
            "min_humidity": 40,
        }
        assert isinstance(statuses[MOCK_DEVICE_HE], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_HE].BreathMode == "normal"
        assert statuses[MOCK_DEVICE_HE].status == {
            "DeviceType": "HE",
            "BreathMode": "normal",
        }

        # Test legacy status
        statuses = api.get_status(legacy=True)
        assert isinstance(statuses[MOCK_DEVICE_AC], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_AC].air_speed == "high"
        assert statuses[MOCK_DEVICE_AC].status == {
            "DeviceType": "AC",
            "air_speed": "high",
            "target_temp": 26,
            "max_temp": 32,
            "min_temp": 16,
        }
        assert isinstance(statuses[MOCK_DEVICE_DH], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_DH].mode == "air_purify"
        assert statuses[MOCK_DEVICE_DH].status == {
            "DeviceType": "DH",
            "mode": "air_purify",
            "max_humidity": 70,
            "min_humidity": 40,
        }
        assert isinstance(statuses[MOCK_DEVICE_HE], JciHitachiAWSStatus)
        assert statuses[MOCK_DEVICE_HE].BreathMode == "normal"
        assert statuses[MOCK_DEVICE_HE].status == {
            "DeviceType": "HE",
            "BreathMode": "normal",
        }

        # Test not existing thing
        statuses = api.get_status("NON_EXISTING_NAME")
        assert statuses == {}

        # Test unknown device type
        with patch.object(api._things[MOCK_DEVICE_AC], "_json") as mock_thing_json:
            mock_thing_json["DeviceType"] = "NON_EXISTING_TYPE"
            statuses = api.get_status(MOCK_DEVICE_AC)
            assert statuses == {}

    def test_refresh_status(self, fixture_aws_mock_api, fixture_aws_identity):
        api = fixture_aws_mock_api
        api._aws_identity = fixture_aws_identity

        thing_name = api.things[MOCK_DEVICE_AC].thing_name
        with patch.object(api, "_mqtt") as mock_mqtt:
            mock_mqtt.publish.return_value = None
            mock_mqtt.execute.return_value = [
                [thing_name],
                [thing_name],
                [thing_name],
                [],
            ]
            mock_mqtt.mqtt_events.mqtt_error_event.is_set.return_value = False

            mock_mqtt.mqtt_events.device_status = {thing_name: ""}
            mock_mqtt.mqtt_events.device_support = {thing_name: ""}
            mock_mqtt.mqtt_events.device_shadow = {thing_name: ""}
            api.refresh_status(
                MOCK_DEVICE_AC, refresh_support_code=True, refresh_shadow=True
            )

            # status event timeout
            mock_mqtt.execute.return_value = [
                [thing_name],
                [thing_name],
                [BaseException],
                [],
            ]
            with pytest.raises(
                RuntimeError,
                match=f"Timed out refreshing {MOCK_DEVICE_AC} status code. Please ensure the device is online and avoid opening the official app.",
            ):
                api.refresh_status(MOCK_DEVICE_AC)

            # status not stored in dict
            mock_mqtt.execute.return_value = [
                [thing_name],
                [thing_name],
                [thing_name],
                [],
            ]
            mock_mqtt.mqtt_events.device_status = {}
            with pytest.raises(
                RuntimeError,
                match=f"An event occurred but wasn't accompanied with data when refreshing {MOCK_DEVICE_AC} status code.",
            ):
                api.refresh_status(MOCK_DEVICE_AC)

            # support event timeout
            mock_mqtt.execute.return_value = [
                [BaseException],
                [thing_name],
                [thing_name],
                [],
            ]
            with pytest.raises(
                RuntimeError,
                match=f"Timed out refreshing {MOCK_DEVICE_AC} support code. Please ensure the device is online and avoid opening the official app.",
            ):
                api.refresh_status(MOCK_DEVICE_AC, refresh_support_code=True)

            # support not stored in dict
            mock_mqtt.execute.return_value = [
                [thing_name],
                [thing_name],
                [thing_name],
                [],
            ]
            mock_mqtt.mqtt_events.device_support = {}
            with pytest.raises(
                RuntimeError,
                match=f"An event occurred but wasn't accompanied with data when refreshing {MOCK_DEVICE_AC} support code.",
            ):
                api.refresh_status(MOCK_DEVICE_AC, refresh_support_code=True)

            # shadow event timeout
            mock_mqtt.execute.return_value = [
                [thing_name],
                [BaseException],
                [thing_name],
                [],
            ]
            with pytest.raises(
                RuntimeError,
                match=f"Timed out refreshing {MOCK_DEVICE_AC} shadow. Please ensure the device is online and avoid opening the official app.",
            ):
                api.refresh_status(MOCK_DEVICE_AC, refresh_shadow=True)

            # shadow not stored in dict
            mock_mqtt.execute.return_value = [
                [thing_name],
                [thing_name],
                [thing_name],
                [],
            ]
            mock_mqtt.mqtt_events.device_shadow = {}
            with pytest.raises(
                RuntimeError,
                match=f"An event occurred but wasn't accompanied with data when refreshing {MOCK_DEVICE_AC} shadow.",
            ):
                api.refresh_status(MOCK_DEVICE_AC, refresh_shadow=True)

    def test_set_status(self, fixture_aws_mock_api, fixture_aws_identity):
        api = fixture_aws_mock_api
        api._aws_identity = fixture_aws_identity

        thing_name = api.things[MOCK_DEVICE_AC].thing_name
        with patch.object(api, "_mqtt") as mock_mqtt:
            mock_mqtt.publish.return_value = None
            mock_mqtt.execute.return_value = [[], [], [], [thing_name]]
            mock_mqtt.mqtt_events.mqtt_error_event.is_set.return_value = False
            mock_mqtt.mqtt_events.device_control.get.return_value = {"FanSpeed": 3}

            assert api.set_status(
                "FanSpeed", device_name=MOCK_DEVICE_AC, status_value=3
            )
            assert api.things[MOCK_DEVICE_AC].status_code.FanSpeed == "moderate"

            # Test legacy name
            assert api.set_status(
                "air_speed", device_name=MOCK_DEVICE_AC, status_value=3
            )
            # Test invalid status_name
            assert not api.set_status(
                "invalid", device_name=MOCK_DEVICE_AC, status_value=3
            )
            # Test invalid status_value
            assert not api.set_status(
                "FanSpeed", device_name=MOCK_DEVICE_AC, status_value=8
            )
            # Test status_str_value
            assert api.set_status(
                "FanSpeed", device_name=MOCK_DEVICE_AC, status_str_value="moderate"
            )
            # Test invalid status_str_value
            assert not api.set_status(
                "FanSpeed", device_name=MOCK_DEVICE_AC, status_str_value="invalid"
            )

            mock_mqtt.mqtt_events.device_control_event.wait.return_value = False
            assert not api.set_status(
                "target_temp", device_name=MOCK_DEVICE_AC, status_value=25
            )

    def test_refresh_monthly_data(self, fixture_aws_mock_api):
        api = fixture_aws_mock_api
        with patch(
            "JciHitachi.aws_connection.GetAvailableAggregationMonthlyData.get_data"
        ) as mock_get_data:
            current_time = time.time()
            mock_get_data.return_value = (
                "OK",
                {
                    "results": {
                        "Data": [
                            {"Timestamp": current_time + 300},
                            {"Timestamp": current_time},
                        ]
                    }
                },
            )

            assert api.things[MOCK_DEVICE_AC].monthly_data is None
            api.refresh_monthly_data(2, MOCK_DEVICE_AC)
            assert api.things[MOCK_DEVICE_AC].monthly_data == [
                {"Timestamp": current_time},
                {"Timestamp": current_time + 300},
            ]

            mock_get_data.return_value = ("Not OK", {})

            with pytest.raises(
                RuntimeError,
                match="An error occurred when getting monthly data: Not OK",
            ):
                api.refresh_monthly_data(2, MOCK_DEVICE_AC)


class TestAWSThing:
    def test_repr(
        self,
        fixture_aws_mock_ac_thing,
        fixture_aws_mock_dh_thing,
        fixture_aws_mock_he_thing,
    ):
        ac_thing = fixture_aws_mock_ac_thing
        assert (
            ac_thing.__repr__()
            == f"""name: {MOCK_DEVICE_AC}
brand: HITACHI
model: RAD-90NF
type: AC
firmware_code: unsupported
firmawre_version: 6.0.035
available: True
status_code: {{'DeviceType': 'AC', 'FanSpeed': 'high', 'TemperatureSetting': 26}}
support_code: {{'DeviceType': 'AC', 'Model': 'RAD-90NF', 'FirmwareVersion': '6.0.035', 'FanSpeed': 31, 'TemperatureSetting': 4128, 'Brand': 'HITACHI', 'max_temp': 32, 'min_temp': 16}}
shadow: None
gateway_mac_address: {MOCK_GATEWAY_MAC}"""
        )

        dh_thing = fixture_aws_mock_dh_thing
        assert (
            dh_thing.__repr__()
            == f"""name: {MOCK_DEVICE_DH}
brand: HITACHI
model: RD-360HH
type: DH
firmware_code: unsupported
firmawre_version: 6.0.035
available: True
status_code: {{'DeviceType': 'DH', 'Mode': 'air_purify'}}
support_code: {{'DeviceType': 'DH', 'Mode': 31, 'FirmwareVersion': '6.0.035', 'Model': 'RD-360HH', 'Brand': 'HITACHI', 'max_humidity': 70, 'min_humidity': 40}}
shadow: None
gateway_mac_address: {MOCK_GATEWAY_MAC}"""
        )

        he_thing = fixture_aws_mock_he_thing
        assert (
            he_thing.__repr__()
            == f"""name: {MOCK_DEVICE_HE}
brand: HITACHI
model: KPI-H
type: HE
firmware_code: unsupported
firmawre_version: 6.0.036
available: True
status_code: {{'DeviceType': 'HE', 'BreathMode': 'normal'}}
support_code: {{'DeviceType': 'HE', 'BreathMode': 7, 'FirmwareVersion': '6.0.036', 'Model': 'KPI-H', 'Brand': 'HITACHI'}}
shadow: None
gateway_mac_address: {MOCK_GATEWAY_MAC}"""
        )

    def test_properties_and_setters(
        self,
        fixture_aws_mock_ac_thing,
        fixture_aws_mock_dh_thing,
        fixture_aws_mock_he_thing,
    ):
        for thing, device_type in (
            (fixture_aws_mock_ac_thing, "AC"),
            (fixture_aws_mock_dh_thing, "DH"),
            (fixture_aws_mock_he_thing, "HE"),
        ):
            # by default (with a setter)
            assert thing.available == True
            assert thing.shadow == None
            assert thing.status_code.DeviceType == device_type
            assert thing.support_code.DeviceType == device_type
            assert thing.monthly_data == None

            # by default (with no setter)
            assert thing.picked_thing == thing._json

            mock_device_types = {
                "AC": ("DH", 2),
                "DH": ("AC", 1),
                "HE": ("AC", 1),
            }

            # assign new values
            thing.available = False
            thing.shadow = {}
            new_status_code = JciHitachiAWSStatus(
                {
                    "DeviceType": mock_device_types[device_type][1],  # mock unknown
                }
            )
            thing.status_code = new_status_code
            new_support_code = thing.support_code = JciHitachiAWSStatusSupport(
                {
                    "DeviceType": mock_device_types[device_type][1],  # mock unknown
                }
            )
            thing.support_code = new_support_code
            thing.monthly_data = [{}]

            # test new values
            assert thing.available == False
            assert thing.shadow == {}
            assert thing.status_code.DeviceType == mock_device_types[device_type][0]
            assert thing.support_code.DeviceType == mock_device_types[device_type][0]
            assert thing.monthly_data == [{}]

    def test_from_device_names(self):
        things = AWSThing.from_device_names(MOCK_THINGS_JSON, None)
        assert len(things) == 2

        # test specifying device
        things = AWSThing.from_device_names(MOCK_THINGS_JSON, MOCK_DEVICE_AC)
        assert len(things) == 1
        assert things[MOCK_DEVICE_AC].name == MOCK_DEVICE_AC

        # test device name unavailable
        with pytest.raises(
            AssertionError, match="Some of device_names are not available from the API."
        ):
            things = AWSThing.from_device_names(MOCK_THINGS_JSON, "unknown")



import time
import warnings

import pytest

from JciHitachi.api import JciHitachiAPI, JciHitachiAWSAPI
from JciHitachi.connection import JciHitachiConnection
from JciHitachi.mqtt_connection import JciHitachiMqttConnection
from JciHitachi.status import JciHitachiAC, JciHitachiDH

from . import (
    MOCK_COMMAND_AC,
    MOCK_COMMAND_DH,
    TEST_DEVICE_AC,
    TEST_DEVICE_DH,
    TEST_EMAIL,
    TEST_PASSWORD,
)


pytestmark = pytest.mark.skipif(
    TEST_EMAIL is None, reason="No secrets for integration tests"
)


@pytest.fixture(scope="session")
def fixture_api():
    api = JciHitachiAPI(
        TEST_EMAIL,
        TEST_PASSWORD,
    )
    api.login()
    return api


@pytest.fixture(scope="session")
def fixture_aws_api():
    api = JciHitachiAWSAPI(
        TEST_EMAIL,
        TEST_PASSWORD,
    )
    api.login()
    return api


@pytest.fixture(scope="function")
def fixture_mqtt(fixture_api):
    mqtt = JciHitachiMqttConnection(
        TEST_EMAIL,
        TEST_PASSWORD,
        fixture_api.user_id,
    )
    return mqtt


class TestAPILogin:
    def test_api(self, fixture_api):
        assert len(fixture_api._session_token) == 31
        # No peripheral available as we have switched to the new API.
        # assert fixture_api._peripherals[TEST_DEVICE_AC].name == TEST_DEVICE_AC

    def test_version(self):
        connection = JciHitachiConnection(TEST_EMAIL, TEST_PASSWORD)
        assert (
            connection._login_response["results"]["LatestVersion"]["UpdateSuggestion"]
            == 0
        )

    def test_session_expiry(self, fixture_api):
        invalid_token = "0000000000000000000000000000000"
        fixture_api._session_token = invalid_token
        fixture_api.login()

        assert len(fixture_api._session_token) == 31
        assert fixture_api._session_token != invalid_token

    def test_deprecated_warning(self):
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            api = JciHitachiAPI(TEST_EMAIL, TEST_PASSWORD)
            # Verify some things
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "This API has been deprecated." in str(w[-1].message)

    @pytest.mark.parametrize("mock_device_name", ["NON_EXISTING_NAME"])
    def test_device_name(self, mock_device_name):
        api = JciHitachiAPI(TEST_EMAIL, TEST_PASSWORD, mock_device_name)

        with pytest.raises(
            AssertionError,
            match=r"Some of device_names are not available from the API.",
        ):
            api.login()

    def test_incorrect_credential(self):
        api = JciHitachiAPI("abc@abc.com", "password")

        with pytest.raises(RuntimeError, match=r"An error occurred when API login:"):
            api.login()


class TestAWSAPILogin:
    def test_api(self, fixture_aws_api):
        assert fixture_aws_api._aws_tokens is not None
        assert fixture_aws_api._aws_identity is not None
        assert fixture_aws_api.things[TEST_DEVICE_AC].name == TEST_DEVICE_AC

    def test_session_expiry(self, fixture_aws_api):
        # Test AWSTokens expiration
        current_access_token = fixture_aws_api._aws_tokens.access_token
        expiration = time.time() + 150.0
        fixture_aws_api._aws_tokens.expiration = expiration
        fixture_aws_api.refresh_status()
        assert fixture_aws_api._aws_tokens.access_token != current_access_token
        assert fixture_aws_api._aws_tokens.expiration != expiration

    @pytest.mark.parametrize("mock_device_name", ["NON_EXISTING_NAME"])
    def test_device_name(self, mock_device_name):
        api = JciHitachiAWSAPI(TEST_EMAIL, TEST_PASSWORD, mock_device_name)

        with pytest.raises(
            AssertionError,
            match=r"Some of device_names are not available from the API.",
        ):
            api.login()

    def test_incorrect_credential(self):
        api = JciHitachiAWSAPI("abc@abc.com", "password")

        with pytest.raises(
            RuntimeError,
            match=r"An error occurred when signing into AWS Cognito Service:",
        ):
            api.login()


class TestMqttLogin:
    def test_mqtt(self, fixture_mqtt):
        fixture_mqtt.configure()
        fixture_mqtt.connect()
        fixture_mqtt.disconnect()

    def test_incorrect_credential(self, fixture_mqtt):
        fixture_mqtt._password = "password"
        fixture_mqtt.configure()
        fixture_mqtt.connect()


class TestACStatus:
    @pytest.mark.slow("online test is a slow test.")
    def test_online(self, fixture_api):
        # Change sound prompt
        current_state = fixture_api.get_status()[TEST_DEVICE_AC]._status[
            JciHitachiAC.idx[MOCK_COMMAND_AC]
        ]
        if current_state != 1:
            changed_state = 1
        else:
            changed_state = 0
        assert fixture_api.set_status(MOCK_COMMAND_AC, changed_state, TEST_DEVICE_AC)

        fixture_api.refresh_status()
        assert (
            fixture_api.get_status()[TEST_DEVICE_AC]._status[
                JciHitachiAC.idx[MOCK_COMMAND_AC]
            ]
            == changed_state
        )

        # Change sound prompt back
        assert fixture_api.set_status(MOCK_COMMAND_AC, current_state, TEST_DEVICE_AC)

        fixture_api.refresh_status()
        assert (
            fixture_api.get_status()[TEST_DEVICE_AC]._status[
                JciHitachiAC.idx[MOCK_COMMAND_AC]
            ]
            == current_state
        )


class TestDHStatus:
    @pytest.mark.skip("Skip online test as no usable account to test.")
    def test_online(self, fixture_api):
        # Change sound control
        current_state = fixture_api.get_status()[TEST_DEVICE_DH]._status[
            JciHitachiDH.idx[MOCK_COMMAND_DH]
        ]
        if current_state != 1:
            changed_state = 1
        else:
            changed_state = 0
        assert fixture_api.set_status(MOCK_COMMAND_DH, changed_state, TEST_DEVICE_DH)

        fixture_api.refresh_status()
        assert (
            fixture_api.get_status()[TEST_DEVICE_DH]._status[
                JciHitachiDH.idx[MOCK_COMMAND_DH]
            ]
            == changed_state
        )

        # Change Change sound control back
        assert fixture_api.set_status(MOCK_COMMAND_DH, current_state, TEST_DEVICE_DH)

        fixture_api.refresh_status()
        assert (
            fixture_api.get_status()[TEST_DEVICE_DH]._status[
                JciHitachiDH.idx[MOCK_COMMAND_DH]
            ]
            == current_state
        )



from functools import lru_cache


class JciHitachiStatus:  # pragma: no cover
    idx = {}

    def __init__(self, status, default) -> None:
        self._status = status
        self._default = default

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return dict((key, getattr(self, key)) for key in self.idx)


class JciHitachiAC(JciHitachiStatus):  # pragma: no cover
    """Data class representing air conditioner status.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {
        "power": 0,
        "mode": 1,
        "air_speed": 2,
        "target_temp": 3,
        "indoor_temp": 4,
        "sleep_timer": 6,
        "vertical_wind_swingable": 14,
        "vertical_wind_direction": 15,
        "horizontal_wind_direction": 17,
        "mold_prev": 23,
        "fast_op": 26,
        "energy_save": 27,
        "sound_prompt": 30,
        "outdoor_temp": 33,
        "power_kwh": 40,
        "freeze_clean": 57,
    }

    def __init__(self, status, default=-1):
        super().__init__(status, default)

    @property
    def power(self):
        """Power. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["power"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def mode(self):
        """Mode. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "cool", "dry", "fan", "auto", "heat", "unknown").
        """

        v = self._status.get(self.idx["mode"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "cool"
        elif v == 1:
            return "dry"
        elif v == 2:
            return "fan"
        elif v == 3:
            return "auto"
        elif v == 4:
            return "heat"
        else:
            return "unknown"

    @property
    def air_speed(self):
        """Air speed. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "silent", "low", "moderate", "high", "unknown").
        """

        v = self._status.get(self.idx["air_speed"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "silent"
        elif v == 2:
            return "low"
        elif v == 3:
            return "moderate"
        elif v == 4:
            return "high"
        else:
            return "unknown"

    @property
    def target_temp(self):
        """Target temperature. Controllable.

        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx["target_temp"], self._default)
        return v

    @property
    def indoor_temp(self):
        """Indoor temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx["indoor_temp"], self._default)
        return v

    @property
    def max_temp(self):
        """Maximum target temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        return 32

    @property
    def min_temp(self):
        """Minimum target temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        return 16

    @property
    def sleep_timer(self):
        """Sleep timer. Controllable.

        Returns
        -------
        int
            Sleep timer (hours).
        """

        v = self._status.get(self.idx["sleep_timer"], self._default)
        return v

    @property
    def vertical_wind_swingable(self):
        """Vertical wind swingable. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["vertical_wind_swingable"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def vertical_wind_direction(self):
        """Vertical wind direction. Controllable.

        Returns
        -------
        int
            Value between 0 to 15.
        """

        v = self._status.get(self.idx["vertical_wind_direction"], self._default)
        return v

    @property
    def horizontal_wind_direction(self):
        """Horizontal wind direction. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "leftmost", "middleleft", "central", "middleright", "rightmost", "unknown").
        """

        v = self._status.get(self.idx["horizontal_wind_direction"], self._default)

        if v > 0:
            v = 6 - v

        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "leftmost"
        elif v == 2:
            return "middleleft"
        elif v == 3:
            return "central"
        elif v == 4:
            return "middleright"
        elif v == 5:
            return "rightmost"
        else:
            return "unknown"

    @property
    def mold_prev(self):
        """Mold prevention. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["mold_prev"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def fast_op(self):
        """Fast operation. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["fast_op"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def energy_save(self):
        """Energy saving. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["energy_save"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def sound_prompt(self):
        """Sound prompt. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "enabled", "disabled", "unknown").
        """

        v = self._status.get(self.idx["sound_prompt"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "enabled"
        elif v == 1:
            return "disabled"
        else:
            return "unknown"

    @property
    def outdoor_temp(self):
        """Outdoor temperature.

        Returns
        -------
        int
            Celsius temperature.
        """

        v = self._status.get(self.idx["outdoor_temp"], self._default)
        return v

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        float
            Kwh.
        """

        v = self._status.get(self.idx["power_kwh"], self._default)
        if v == -1:
            return v
        return v / 10.0

    @property
    def freeze_clean(self):
        """Freeze clean. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["freeze_clean"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"


class JciHitachiDH(JciHitachiStatus):  # pragma: no cover
    """Data class representing dehumidifier status.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {
        "power": 0,
        "mode": 1,
        "target_humidity": 3,
        "indoor_humidity": 7,
        "wind_swingable": 8,
        "water_full_warning": 10,
        "clean_filter_notify": 11,
        "air_purify_level": 13,
        "air_speed": 14,
        "side_vent": 15,
        "sound_control": 16,
        "error_code": 18,
        "mold_prev": 19,
        "power_kwh": 29,
        "air_quality_value": 35,
        "air_quality_level": 36,
        "pm25_value": 37,
        "display_brightness": 39,
        "odor_level": 40,
        "air_cleaning_filter": 41,
    }

    def __init__(self, status, default=-1):
        super().__init__(status, default)

    @property
    def power(self):
        """Power. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["power"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def mode(self):
        """Mode. Controllable.

        Returns
        -------
        str
            One of (
            "unsupported", "auto", "custom", "continuous", "clothes_dry",
            "air_purify", "mold_prev", "low_humidity", "eco_comfort", "unknown"
            ).
        """

        v = self._status.get(self.idx["mode"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "custom"
        elif v == 2:
            return "continuous"
        elif v == 3:
            return "clothes_dry"
        elif v == 4:
            return "air_purify"
        elif v == 5:
            return "mold_prev"
        elif v == 8:
            return "low_humidity"
        elif v == 9:
            return "eco_comfort"
        else:
            return "unknown"

    @property
    def target_humidity(self):
        """Target humidity. Controllable.

        Returns
        -------
        int
            Relative humidity.
        """

        v = self._status.get(self.idx["target_humidity"], self._default)
        return v

    @property
    def indoor_humidity(self):
        """Indoor humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        v = self._status.get(self.idx["indoor_humidity"], self._default)
        return v

    @property
    def max_humidity(self):
        """Maximum target humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        return 70

    @property
    def min_humidity(self):
        """Minimum target humidity.

        Returns
        -------
        int
            Relative humidity.
        """

        return 40

    @property
    def wind_swingable(self):
        """Wind swingable. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["wind_swingable"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def water_full_warning(self):
        """Water full warning.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["water_full_warning"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"  # not activated
        elif v == 1:
            return "on"  # activated
        else:
            return "unknown"

    @property
    def clean_filter_notify(self):
        """Clean filter notify control. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["clean_filter_notify"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"

    @property
    def air_purify_level(self):
        """Air purify level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"

    @property
    def air_speed(self):
        """Air speed. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "auto", "silent", "low", "moderate", "high", "unknown").
        """

        v = self._status.get(self.idx["air_speed"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "auto"
        elif v == 1:
            return "silent"
        elif v == 2:
            return "low"
        elif v == 3:
            return "moderate"
        elif v == 4:
            return "high"
        else:
            return "unknown"

    @property
    def side_vent(self):
        """Side vent.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["side_vent"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def sound_control(self):
        """Sound control. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "silent", "button", "button+waterfull", "unknown").
        """

        v = self._status.get(self.idx["sound_control"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "silent"
        elif v == 1:
            return "button"
        elif v == 2:
            return "button+waterfull"
        else:
            return "unknown"

    @property
    def error_code(self):
        """Error code.

        Returns
        -------
        int
            Error code.
        """

        v = self._status.get(self.idx["error_code"], self._default)
        return v

    @property
    def mold_prev(self):
        """Mold prevention. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "off", "on", "unknown").
        """

        v = self._status.get(self.idx["mold_prev"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "off"
        elif v == 1:
            return "on"
        else:
            return "unknown"

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        float
            Kwh.
        """

        v = self._status.get(self.idx["power_kwh"], self._default)
        if v == -1:
            return v
        return v / 10.0

    @property
    def air_quality_value(self):
        """Air quality value. Not implemented.

        Returns
        -------
        int
            Not implemented.
        """
        return self._default

    @property
    def air_quality_level(self):
        """Air quality level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"

    @property
    def pm25_value(self):
        """PM2.5 value.

        Returns
        -------
        int
            PM2.5 value.
        """

        v = self._status.get(self.idx["pm25_value"], self._default)
        return v

    @property
    def display_brightness(self):
        """Display brightness. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "bright", "dark", "off", "all_off" "unknown").
        """
        v = self._status.get(self.idx["display_brightness"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "bright"
        elif v == 1:
            return "dark"
        elif v == 2:
            return "off"
        elif v == 3:
            return "all_off"
        else:
            return "unknown"

    @property
    def odor_level(self):
        """Odor level.

        Returns
        -------
        str
            One of ("unsupported", "low", "middle", "high", "unknown").
        """
        v = self._status.get(self.idx["odor_level"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "low"
        elif v == 1:
            return "middle"
        elif v == 2:
            return "high"
        else:
            return "unknown"

    @property
    def air_cleaning_filter(self):
        """Air cleaning filter setting.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["air_cleaning_filter"], self._default)
        if v == -1:
            return "unsupported"
        elif v == 0:
            return "disabled"
        elif v == 1:
            return "enabled"
        else:
            return "unknown"


class JciHitachiHE(JciHitachiStatus):  # pragma: no cover
    """Data class representing heat exchanger status. Not implemented.

    Parameters
    ----------
    status : dict
        Status retrieved from JciHitachiStatusInterpreter.decode_status().
    default : int, optional
        Default value when a status doesn't exist, by default -1.
    """

    idx = {}

    def __init__(self, status, default=-1):
        super().__init__(status, default)


class JciHitachiStatusSupport(JciHitachiStatus):  # pragma: no cover
    supported_type = {}

    def __init__(self, status, default=0):
        super().__init__(status, default)

    @property
    def brand(self):
        """Device brand.

        Returns
        -------
        str
            Device brand.
        """

        v = self._status.get(self.idx["brand"], self._default)
        return v

    @property
    def model(self):
        """Device model.

        Returns
        -------
        str
            Device model.
        """

        v = self._status.get(self.idx["model"], self._default)
        return v

    def _uni_v(self, v):
        return (((v >> 0x10) & 0xFF) << 8) | (v >> 0x18)

    def _dual_v(self, v):
        low = (v >> 0x10) & 0xFF
        high = (v >> 0x18) & 0xFF
        return low, high

    def _functional_v(self, v):
        uni_v = self._uni_v(v)
        return ((uni_v >> i) & 0x1 for i in range(16))

    def limit(self, status_name, status_value):
        """Limit status_value within an acceptable range.

        Parameters
        ----------
        status_name : str
            Status name, which has to be in idx dict. E.g. JciHitachiAC.idx
        status_value : int
            Status value.

        Returns
        -------
        int or None
            If the status_value can be limited with an acceptable raneg, return int.
            Otherwise, if the status_value is invalid, return None.
        """

        is_support, *v = getattr(self, status_name)
        if not is_support:
            return None

        supported_type = self.supported_type[status_name]
        if supported_type == "uni":
            return min(status_value, v[0])
        elif supported_type == "dual":
            return min(v[1], max(status_value, v[0]))
        elif supported_type == "functional":
            if v[status_value]:
                return status_value
        return None


class JciHitachiACSupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported air conditioner status.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {
        "brand": "brand",
        "model": "model",
        "power": 0,
        "mode": 1,
        "air_speed": 2,
        "target_temp": 3,
        "indoor_temp": 4,
        "sleep_timer": 6,
        "vertical_wind_swingable": 14,
        "vertical_wind_direction": 15,
        "horizontal_wind_direction": 17,
        "mold_prev": 23,
        "fast_op": 26,
        "energy_save": 27,
        "sound_prompt": 30,
        "outdoor_temp": 33,
        "power_kwh": 40,
    }

    supported_type = {
        "brand": "str",
        "model": "str",
        "power": "functional",
        "mode": "functional",
        "air_speed": "functional",
        "target_temp": "dual",
        "indoor_temp": "dual",
        "sleep_timer": "uni",
        "vertical_wind_swingable": "functional",
        "vertical_wind_direction": "functional",
        "horizontal_wind_direction": "functional",
        "mold_prev": "functional",
        "fast_op": "functional",
        "energy_save": "functional",
        "sound_prompt": "functional",
        "outdoor_temp": "dual",
        "power_kwh": "uni",
    }

    def __init__(self, status, default=0):
        super().__init__(status, default)

    @property
    def power(self):
        """Power. Controllable.

        Returns
        -------
        (bool, int, int)
            (is_support, off, on).
        """

        v = self._status.get(self.idx["power"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def mode(self):
        """Mode. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, (cool, dry, fan, auto, heat, 0...).
        """

        v = self._status.get(self.idx["mode"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def air_speed(self):
        """Air speed. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "silent", "low", "moderate", "high", 0...).
        """

        v = self._status.get(self.idx["air_speed"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def target_temp(self):
        """Target temperature. Controllable.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx["target_temp"], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def indoor_temp(self):
        """Indoor temperature.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx["indoor_temp"], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def sleep_timer(self):
        """Sleep timer. Controllable.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx["sleep_timer"], self._default)
        support = self._uni_v(v)

        return (v != 0, support)

    @property
    def vertical_wind_swingable(self):
        """Vertical wind swingable. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx["vertical_wind_swingable"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def vertical_wind_direction(self):
        """Vertical wind direction. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "level1", "level2", "level3", "level4", "level5", "level6", "level7", 0...).
        """

        v = self._status.get(self.idx["vertical_wind_direction"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def horizontal_wind_direction(self):
        """Horizontal wind direction. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "leftmost", "middleleft", "central", "middleright", "rightmost", 0...).
        """

        v = self._status.get(self.idx["horizontal_wind_direction"], self._default)
        if v > 0:
            v = 6 - v

        supports = self._functional_v(v)
        return (v != 0, *supports)

    @property
    def mold_prev(self):
        """Mold prevention. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx["mold_prev"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def fast_op(self):
        """Fast operation. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("disabled", "enabled", 0...).
        """

        v = self._status.get(self.idx["fast_op"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def energy_save(self):
        """Energy saving. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx["energy_save"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def sound_prompt(self):
        """Sound prompt. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...).
        """

        v = self._status.get(self.idx["sound_prompt"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def outdoor_temp(self):
        """Outdoor temperature.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx["outdoor_temp"], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx["power_kwh"], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)


class JciHitachiDHSupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported dehumidifier status.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {
        "brand": "brand",
        "model": "model",
        "power": 0,
        "mode": 1,
        "target_humidity": 3,
        "indoor_humidity": 7,
        "wind_swingable": 8,
        "water_full_warning": 10,
        "clean_filter_notify": 11,
        "air_purify_level": 13,
        "air_speed": 14,
        "side_vent": 15,
        "sound_control": 16,
        "error_code": 18,
        "mold_prev": 19,
        "power_kwh": 29,
        "air_quality_value": 35,
        "air_quality_level": 36,
        "pm25_value": 37,
        "display_brightness": 39,
        "odor_level": 40,
        "air_cleaning_filter": 41,
    }

    supported_type = {
        "brand": "str",
        "model": "str",
        "power": "functional",
        "mode": "functional",
        "target_humidity": "dual",
        "indoor_humidity": "dual",
        "wind_swingable": "functional",
        "water_full_warning": "functional",
        "clean_filter_notify": "functional",
        "air_purify_level": "functional",  # not implemented
        "air_speed": "functional",
        "side_vent": "functional",
        "sound_control": "functional",
        "error_code": "uni",
        "mold_prev": "functional",
        "power_kwh": "uni",
        "air_quality_value": "uni",
        "air_quality_level": "functional",
        "pm25_value": "uni",
        "display_brightness": "functional",
        "odor_level": "functional",
        "air_cleaning_filter": "functional",
    }

    def __init__(self, status, default=0):
        super().__init__(status, default)

    @property
    def power(self):
        """Power. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            (is_support, off, on).
        """

        v = self._status.get(self.idx["power"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def mode(self):
        """Mode. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, (auto, custom, continuous, clothes_dry, air_purify, mold_prev, air_supply, human_comfort, low_humidity, eco_comfort, 0...).
        """

        v = self._status.get(self.idx["mode"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def target_humidity(self):
        """Target humidity. Controllable.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx["target_humidity"], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def indoor_humidity(self):
        """Indoor humidity.

        Returns
        -------
        (bool, int, int)
            (is_support, minimum, maximum)
        """

        v = self._status.get(self.idx["indoor_humidity"], self._default)
        supports = self._dual_v(v)

        return (v != 0, *supports)

    @property
    def wind_swingable(self):
        """Wind swingable. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """

        v = self._status.get(self.idx["wind_swingable"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def water_full_warning(self):
        """Water full warning.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """

        v = self._status.get(self.idx["water_full_warning"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def clean_filter_notify(self):
        """Clean filter notify control. Controllable.

        Returns
        -------
        str
            One of ("unsupported", "disabled", "enabled", "unknown").
        """

        v = self._status.get(self.idx["clean_filter_notify"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def air_purify_level(self):
        """Air purify level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """

        return "unsupported"

    @property
    def air_speed(self):
        """Air speed. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("auto", "silent", "low", "moderate", "high", 0...).
        """

        v = self._status.get(self.idx["air_speed"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def side_vent(self):
        """Side vent.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("off", "on", 0...).
        """

        v = self._status.get(self.idx["side_vent"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def sound_control(self):
        """Sound control. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("silent", "button", "button+waterfull", 0...).
        """

        v = self._status.get(self.idx["sound_control"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def error_code(self):
        """Error code.

        Returns
        -------
        (bool,)
            (is_support,).
        """

        v = self._status.get(self.idx["error_code"], self._default)

        return (v != 0,)

    @property
    def mold_prev(self):
        """Mold prevention. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("silent", "off", "on",, 0...).
        """

        v = self._status.get(self.idx["mold_prev"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def power_kwh(self):
        """Accumulated Kwh in a day.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx["power_kwh"], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)

    @property
    def air_quality_value(self):
        """Air quality value. Not implemented.

        Returns
        -------
        int
            Not implemented.
        """
        return self._default

    @property
    def air_quality_level(self):
        """Air quality level. Not implemented.

        Returns
        -------
        str
            Not implemented.
        """
        return "unsupported"

    @property
    def pm25_value(self):
        """PM2.5 value.

        Returns
        -------
        (bool, int)
            (is_support, maximum).
        """

        v = self._status.get(self.idx["pm25_value"], self._default)
        supports = self._uni_v(v)

        return (v != 0, supports)

    @property
    def display_brightness(self):
        """Display brightness. Controllable.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("bright", "dark", "off", 0...).
        """

        v = self._status.get(self.idx["display_brightness"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def odor_level(self):
        """Odor level.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("low", "middle", "high", 0...).
        """

        v = self._status.get(self.idx["odor_level"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)

    @property
    def air_cleaning_filter(self):
        """Air cleaning filter setting.

        Returns
        -------
        (bool, Tuple[int])
            is_support, ("enabled", "disabled", 0...). status's reversed order.
        """

        v = self._status.get(self.idx["air_cleaning_filter"], self._default)
        supports = self._functional_v(v)

        return (v != 0, *supports)


class JciHitachiHESupport(JciHitachiStatusSupport):  # pragma: no cover
    """Data model representing supported heat exchanger status. Not implemented.

    Parameters
    ----------
    status : dict
        Supported status retrieved from JciHitachiStatusInterpreter.decode_support().
    default : int, optional
        Default value when a status doesn't exist, by default 0.
    """

    idx = {}

    def __init__(self, status, default=0):
        super().__init__(status, default)


STATUS_DICT = {
    "AC": {
        "DeviceType": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            },
        },
        "Switch": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "power",
            "id2str": {
                0: "off",
                1: "on",
            },
        },
        "Mode": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "mode",
            "id2str": {
                0: "cool",
                1: "dry",
                2: "fan",
                3: "auto",
                4: "heat",
            },
        },
        "FanSpeed": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "air_speed",
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
                5: "rapid",
                6: "express",
            },
        },
        "TemperatureSetting": {
            "controllable": True,
            "is_numeric": True,
            "legacy_name": "target_temp",
        },
        "IndoorTemperature": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "indoor_temp",
        },
        "SleepModeRemainingTime": {
            "controllable": True,
            "is_numeric": True,
            "legacy_name": "sleep_timer",
        },
        "VerticalWindDirectionSwitch": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "vertical_wind_swingable",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "VerticalWindDirectionSetting": {
            "controllable": True,
            "is_numeric": True,
            "legacy_name": "vertical_wind_direction",
        },
        "HorizontalWindDirectionSetting": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "horizontal_wind_direction",
            "id2str": {
                0: "auto",
                1: "leftmost",
                2: "middleleft",
                3: "central",
                4: "middleright",
                5: "rightmost",
            },
        },
        "MildewProof": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "mold_prev",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "QuickMode": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "fast_op",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "PowerSaving": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "energy_save",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "ControlTone": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "sound_prompt",
            "id2str": {
                0: "enabled",
                1: "disabled",
            },
        },
        "PowerConsumption": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "power_kwh",
        },
        "TaiseiaError": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "FilterElapsedHour": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "CleanSwitch": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "freeze_clean",
            "id2str": {
                0: "off",
                1: "on",
            },
        },
        "CleanNotification": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "CleanStatus": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "Error": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "max_temp": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "max_temp",
        },
        "min_temp": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "min_temp",
        },
        "Panel": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "bright",
                1: "dark",
                2: "off",
                3: "all_off",
            },
        },
    },
    "DH": {
        "DeviceType": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            },
        },
        "Switch": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "power",
            "id2str": {
                0: "off",
                1: "on",
            },
        },
        "Mode": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "mode",
            "id2str": {
                0: "auto",
                1: "custom",
                2: "continuous",
                3: "clothes_dry",
                4: "air_purify",
                5: "mold_prev",
                8: "low_humidity",
                9: "eco_comfort",
            },
        },
        "FanSpeed": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "air_speed",
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
            },
        },
        "MildewProof": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "mold_prev",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "ControlTone": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "sound_control",
            "id2str": {
                0: "silent",
                1: "button",
                2: "button+waterfull",
            },
        },
        "SaaControlTone": {  # currently not supported
            "controllable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {},
        },
        "PowerConsumption": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "power_kwh",
        },
        "Ion": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "HumiditySetting": {
            "controllable": True,
            "is_numeric": True,
            "legacy_name": "target_humidity",
        },
        "AutoWindDirection": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "wind_swingable",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "KeypadLock": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "DisplayBrightness": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "display_brightness",
            "id2str": {
                0: "bright",
                1: "dark",
                2: "off",
                3: "all_off",
            },
        },
        "FilterControl": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "air_cleaning_filter",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "PM25": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "pm25_value",
        },
        "IndoorHumidity": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "indoor_humidity",
        },
        "SideAirOutlet": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "side_vent",
            "id2str": {
                0: "off",
                1: "on",
            },
        },
        "Defrost": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {},
        },
        "SmellIndex": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "odor_level",
            "id2str": {
                0: "low",
                1: "middle",
                2: "high",
            },
        },
        "CleanFilterNotification": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": "clean_filter_notify",
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "TankFullNotification": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "water_full_warning",
            "id2str": {
                0: "off",  # not activated
                1: "on",  # activated
            },
        },
        "TaiseiaError": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "Error": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "error_code",
        },
        "max_humidity": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "max_humidity",
        },
        "min_humidity": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": "min_humidity",
        },
    },
    "HE": {
        "DeviceType": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": "DeviceType",
            "id2str": {
                1: "AC",
                2: "DH",
                3: "HE",
                4: "PM25_PANEL",
            },
        },
        "Switch": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "off",
                1: "on",
            },
        },
        "Mode": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "air_condition",
                1: "dehumidification",
                2: "air_supply",
                3: "auto",
                4: "heater",
            },
        },
        "FanSpeed": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "auto",
                1: "silent",
                2: "low",
                3: "moderate",
                4: "high",
            },
        },
        "IndoorTemperature": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "TaiseiaError": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
        "CleanFilterNotification": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "BreathMode": {
            "controllable": True,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {0: "auto", 1: "energy_recovery", 2: "normal"},
        },
        "FrontFilterNotification": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "Pm25FilterNotification": {
            "controllable": False,
            "is_numeric": False,
            "legacy_name": None,
            "id2str": {
                0: "disabled",
                1: "enabled",
            },
        },
        "Error": {
            "controllable": False,
            "is_numeric": True,
            "legacy_name": None,
        },
    },
    "PM25_PANEL": {},
}


class JciHitachiAWSStatus:
    """Data class representing `AWSThing` status.

    Parameters
    ----------
    raw_status : dict
        Status retrieved from `JciHitachiAWSMqttConnection` _on_publish() callback.
    legacy : bool, optional
        Whether the raw_status is a legacy status, i.e. derived from subclasses of `JciHitachiStatus`, by default False.
    """

    device_type_mapping = {
        1: "AC",
        2: "DH",
        3: "HE",
        4: "PM25_PANEL",
    }

    def __init__(self, raw_status: dict, legacy=False) -> None:
        self._raw_status: dict = raw_status
        self._status: dict = raw_status if legacy else self._preprocess(raw_status)
        self._device_type: str = self._status["DeviceType"]

    def __getattr__(self, name):
        return self._status.get(name, "unsupported")

    def __repr__(self) -> str:
        return str(self._status)

    def _preprocess(self, raw_status):
        # device type
        if "DeviceType" not in raw_status or (
            "DeviceType" in raw_status
            and raw_status["DeviceType"] not in self.device_type_mapping
        ):
            raise AttributeError(
                "`DeviceType` isn't in the raw status or has an invalid value."
            )

        if "PowerConsumption" in raw_status:
            raw_status["PowerConsumption"] /= 10.0

        status = {}
        device_type = self.device_type_mapping[raw_status["DeviceType"]]
        for key, value in raw_status.items():
            if key in STATUS_DICT[device_type]:
                if STATUS_DICT[device_type][key]["is_numeric"]:
                    status[key] = value
                else:
                    status[key] = STATUS_DICT[device_type][key]["id2str"].get(
                        value, "unknown"
                    )

        return status

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return self._status

    @property
    def legacy_status(self):
        """All legacy status name used by the old API.

        Returns
        -------
        JciHitachiAWSStatus
            Status with legacy name.
        """

        status = {}
        device_type = self._status["DeviceType"]
        for status_name, status_value in self._status.items():
            if status_name in STATUS_DICT[device_type]:
                if STATUS_DICT[device_type][status_name]["legacy_name"] is None:
                    # no legacy name
                    status.update({status_name: status_value})
                else:
                    status.update(
                        {
                            STATUS_DICT[device_type][status_name][
                                "legacy_name"
                            ]: status_value
                        }
                    )
        return JciHitachiAWSStatus(status, legacy=True)

    @staticmethod
    @lru_cache
    def str2id(
        device_type: str,
        status_name: str,
        status_value: int = None,
        status_str_value: str = None,
        support_code: int = None,
    ):
        is_valid = (status_value is not None) ^ (status_str_value is not None)

        # Name check
        if is_valid:
            if status_name not in STATUS_DICT[device_type]:
                legacy2new = {
                    specs["legacy_name"]: new_status_name
                    for new_status_name, specs in STATUS_DICT[device_type].items()
                }
                if status_name in legacy2new:
                    status_name = legacy2new[status_name]
                else:
                    is_valid = False

        # Value check
        if is_valid:
            if status_str_value is not None:
                str2id_dict = {
                    value: key
                    for key, value in STATUS_DICT[device_type][status_name][
                        "id2str"
                    ].items()
                }
                if status_str_value in str2id_dict:
                    status_value = str2id_dict[status_str_value]
                else:
                    is_valid = False
            else:
                if (
                    not STATUS_DICT[device_type][status_name]["is_numeric"]
                    and status_value
                    not in STATUS_DICT[device_type][status_name]["id2str"]
                ):
                    is_valid = False

        # if support_code is specified, we check whether the given status value is valid for the device
        if is_valid and support_code is not None:
            if STATUS_DICT[device_type][status_name]["is_numeric"]:
                if status_value > support_code.status[status_name]:
                    is_valid = False
            else:
                if 2**status_value & support_code.status[status_name] == 0:
                    is_valid = False

        return is_valid, status_name, status_value

    def set_new_status(self, name: str, value: int):
        if STATUS_DICT[self._device_type][name]["is_numeric"]:
            self._status[name] = value
        else:
            self._status[name] = STATUS_DICT[self._device_type][name]["id2str"].get(
                value, "unknown"
            )


class JciHitachiAWSStatusSupport:
    """Data class representing `AWSThing` status support.

    Parameters
    ----------
    raw_status : dict
        Status retrieved from `JciHitachiAWSMqttConnection` _on_publish() callback.
    """

    device_type_mapping = JciHitachiAWSStatus.device_type_mapping

    def __init__(self, raw_status: dict) -> None:
        self._raw_status: dict = raw_status
        self._status: dict = self._preprocess(raw_status)

    def __getattr__(self, name):
        return self._status.get(name, "unsupported")

    def __repr__(self) -> str:
        return str(self._status)

    def _preprocess(self, status):
        status = status.copy()

        if status.get("Error", 0) != 0:
            return status

        # device type
        status["DeviceType"] = self.device_type_mapping[status["DeviceType"]]

        status["Brand"] = "HITACHI"

        if status["DeviceType"] == "AC":
            status["max_temp"] = (
                status["TemperatureSetting"] & 255
                if "TemperatureSetting" in status
                else 32
            )
            status["min_temp"] = (
                status["TemperatureSetting"] >> 8 & 255
                if "TemperatureSetting" in status
                else 16
            )
        elif status["DeviceType"] == "DH":
            status["max_humidity"] = (
                status["HumiditySetting"] & 255 if "HumiditySetting" in status else 70
            )
            status["min_humidity"] = (
                status["HumiditySetting"] >> 8 & 255
                if "HumiditySetting" in status
                else 40
            )

        return status

    @property
    def status(self):
        """All status.

        Returns
        -------
        dict
            All status.
        """

        return self._status



from JciHitachi.api import AWSThing
from JciHitachi.model import (
    STATUS_DICT,
    JciHitachiAWSStatus,
    JciHitachiAWSStatusSupport,
)


class TestStatusDict:
    device_types = AWSThing.supported_device_type.values()

    def test_sanity(self):
        attribute_keys = {"controllable", "is_numeric", "legacy_name"}

        for device_type in self.device_types:
            for status_name, status_attributes in STATUS_DICT[device_type].items():
                # Test if attribute keys are in status attributes.
                assert (
                    len(attribute_keys - set(status_attributes)) == 0
                ), f"Lack of attribute keys: {status_name} {set(status_attributes)}"
                # Test is_numeric and id2str mutual exclusiveness
                assert status_attributes["is_numeric"] ^ (
                    "id2str" in status_attributes
                ), f"is_numeric and id2str are mutually exclusive: {status_name}"


class TestModel:
    def test_ac_correctness(self):
        raw_ac_status = {
            "DeviceType": 1,
            "Switch": 0,
            "Mode": 0,
            "FanSpeed": 3,
            "TemperatureSetting": 26,
            "IndoorTemperature": 23,
            "SleepModeRemainingTime": 0,
            "MildewProof": 0,
            "QuickMode": 0,
            "PowerSaving": 0,
            "ControlTone": 0,
            "PowerConsumption": 5,
            "TaiseiaError": 0,
            "FilterElapsedHour": 800,
            "CleanSwitch": 0,
            "CleanNotification": 0,
            "CleanStatus": 0,
            "Error": 0,
            "RequestTimestamp": 1648618952,
            "Timestamp": 1648618953,
        }
        processed_ac_status = {
            "CleanNotification": 0,
            "CleanStatus": 0,
            "CleanSwitch": "off",
            "ControlTone": "enabled",
            "DeviceType": "AC",
            "Error": 0,
            "FanSpeed": "moderate",
            "FilterElapsedHour": 800,
            "IndoorTemperature": 23,
            "MildewProof": "disabled",
            "Mode": "cool",
            "PowerConsumption": 0.5,
            "PowerSaving": "disabled",
            "QuickMode": "disabled",
            "SleepModeRemainingTime": 0,
            "Switch": "off",
            "TaiseiaError": 0,
            "TemperatureSetting": 26,
        }

        status = JciHitachiAWSStatus(raw_ac_status)
        assert status.status == processed_ac_status
        assert (
            len(processed_ac_status) == len(raw_ac_status) - 2
        )  # no `RequestTimestamp` and `Timestamp`

        raw_ac_support = {
            "FirmwareId": 3,
            "DeviceType": 1,
            "Model": "RAD-90NF",
            "FindMe": 10,
            "Switch": 3,
            "Mode": 31,
            "FanSpeed": 31,
            "TemperatureSetting": 4128,
            "IndoorTemperature": 40,
            "SleepModeRemainingTime": 1440,
            "MildewProof": 3,
            "QuickMode": 3,
            "PowerSaving": 3,
            "ControlTone": 3,
            "PowerConsumption": 65535,
            "TaiseiaError": 85,
            "FilterElapsedHour": 800,
            "CleanSwitch": 3,
            "CleanNotification": 3,
            "CleanStatus": 7,
            "Error": 0,
            "FirmwareVersion": "6.0.035",
            "FirmwareCode": 35,
            "SystemTimestamp": 35000,
            "RequestTimestamp": 1648618952,
            "Timestamp": 1648618952,
        }
        processed_ac_support = {
            "FirmwareId": 3,
            "DeviceType": "AC",
            "Model": "RAD-90NF",
            "FindMe": 10,
            "Switch": 3,
            "Mode": 31,
            "FanSpeed": 31,
            "TemperatureSetting": 4128,
            "IndoorTemperature": 40,
            "SleepModeRemainingTime": 1440,
            "MildewProof": 3,
            "QuickMode": 3,
            "PowerSaving": 3,
            "ControlTone": 3,
            "PowerConsumption": 65535,
            "TaiseiaError": 85,
            "FilterElapsedHour": 800,
            "CleanSwitch": 3,
            "CleanNotification": 3,
            "CleanStatus": 7,
            "Error": 0,
            "FirmwareVersion": "6.0.035",
            "FirmwareCode": 35,
            "SystemTimestamp": 35000,
            "RequestTimestamp": 1648618952,
            "Timestamp": 1648618952,
            "Brand": "HITACHI",
            "max_temp": 32,
            "min_temp": 16,
        }

        support = JciHitachiAWSStatusSupport(raw_ac_support)
        assert support.status == processed_ac_support
        assert (
            len(processed_ac_support) == len(raw_ac_support) + 3
        )  # `Brand` `max_temp` `min_temp`

    def test_he_correctness(self):
        raw_he_status = {
            "DeviceType": 3,
            "Switch": 0,
            "FanSpeed": 4,
            "IndoorTemperature": 31,
            "TaiseiaError": 0,
            "CleanFilterNotification": 0,
            "BreathMode": 0,
            "FrontFilterNotification": 0,
            "Pm25FilterNotification": 0,
            "Error": 0,
            "RequestTimestamp": 1661593958,
            "ReceiveTimestamp": 1661593958,
            "Timestamp": 1661593958,
        }
        processed_he_status = {
            "DeviceType": "HE",
            "Switch": "off",
            "FanSpeed": "high",
            "IndoorTemperature": 31,
            "TaiseiaError": 0,
            "CleanFilterNotification": "disabled",
            "BreathMode": "auto",
            "FrontFilterNotification": "disabled",
            "Pm25FilterNotification": "disabled",
            "Error": 0,
        }

        status = JciHitachiAWSStatus(raw_he_status)
        assert status.status == processed_he_status
        assert (
            len(processed_he_status) == len(raw_he_status) - 3
        )  # no `RequestTimestamp`, `ReceiveTimestamp``, and `Timestamp`

        raw_he_support = {
            "FirmwareId": 1,
            "DeviceType": 3,
            "Model": "KPI-H",
            "FindMe": 10,
            "Switch": 3,
            "FanSpeed": 20,
            "IndoorTemperature": 33151,
            "TaiseiaError": 244,
            "CleanFilterNotification": 1,
            "BreathMode": 7,
            "FrontFilterNotification": 1,
            "Pm25FilterNotification": 1,
            "Error": 0,
            "FirmwareVersion": "6.0.036",
            "FirmwareCode": 52,
            "SystemTimestamp": 81274,
            "RequestTimestamp": 1661596297,
            "Timestamp": 1661596297,
        }
        processed_he_support = {
            "FirmwareId": 1,
            "DeviceType": "HE",
            "Model": "KPI-H",
            "FindMe": 10,
            "Switch": 3,
            "FanSpeed": 20,
            "IndoorTemperature": 33151,
            "TaiseiaError": 244,
            "CleanFilterNotification": 1,
            "BreathMode": 7,
            "FrontFilterNotification": 1,
            "Pm25FilterNotification": 1,
            "Error": 0,
            "FirmwareVersion": "6.0.036",
            "FirmwareCode": 52,
            "SystemTimestamp": 81274,
            "RequestTimestamp": 1661596297,
            "Timestamp": 1661596297,
            "Brand": "HITACHI",
        }

        support = JciHitachiAWSStatusSupport(raw_he_support)
        assert support.status == processed_he_support
        assert len(processed_he_support) == len(raw_he_support) + 1  # `Brand`

    def test_str2id(self):
        raw_ac_support = {
            "FirmwareId": 3,
            "DeviceType": 1,
            "Model": "RAD-90NF",
            "FindMe": 10,
            "Switch": 3,
            "Mode": 31,
            "FanSpeed": 31,
            "TemperatureSetting": 4128,
            "IndoorTemperature": 40,
            "SleepModeRemainingTime": 1440,
            "MildewProof": 3,
            "QuickMode": 3,
            "PowerSaving": 3,
            "ControlTone": 3,
            "PowerConsumption": 65535,
            "TaiseiaError": 85,
            "FilterElapsedHour": 800,
            "CleanSwitch": 3,
            "CleanNotification": 3,
            "CleanStatus": 7,
            "Error": 0,
            "FirmwareVersion": "6.0.035",
            "FirmwareCode": 35,
            "SystemTimestamp": 35000,
            "RequestTimestamp": 1648618952,
            "Timestamp": 1648618952,
        }
        support = JciHitachiAWSStatusSupport(raw_ac_support)

        # test legacy name and string status
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id(
            "AC", "air_speed", status_str_value="high", support_code=support
        )
        assert is_valid == True
        assert status_name == "FanSpeed"
        assert status_value == 4

        # test regular name and integer status
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id(
            "AC", "FanSpeed", 4, support_code=support
        )
        assert is_valid == True
        assert status_name == "FanSpeed"
        assert status_value == 4

        # test incorrect status name
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id(
            "AC", "FanpSpeed", 4, support_code=support
        )
        assert is_valid == False

        # test incorrect status value
        is_valid, status_name, status_value = JciHitachiAWSStatus.str2id(
            "AC", "FanSpeed", 6, support_code=support
        )
        assert is_valid == False



import os
import sys

from JciHitachi import __author__, __version__

git_repo_path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
requirements_txt_path = os.path.join(git_repo_path, "requirements.txt")
requirements_test_txt_path = os.path.join(git_repo_path, "requirements_test.txt")

sys.path.append(git_repo_path)
sys.path.append(os.path.join(git_repo_path, "docs", "source"))

import conf

from setup import install_requires, tests_require


class TestSanity:
    def test_annotions_consistency(self):
        assert __author__ == conf.author
        assert f"v{__version__}" == conf.release

    def test_install_requirements_consistency(self):
        with open(requirements_txt_path, "r", encoding="utf-8") as f:
            req_txt = [req for req in f.read().split("\n") if req[0] != "-"]
        assert set(install_requires) == set(req_txt)

    def test_test_requirements_consistency(self):
        with open(requirements_test_txt_path, "r", encoding="utf-8") as f:
            req_txt = [req for req in f.read().split("\n") if req[0] != "-"]
        assert set(tests_require) == set(req_txt)
